/*
Copyright (c) 2012, Felix J. Ogris
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met: 

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer. 
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <limits.h>
#include <err.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <fcntl.h>
#include <signal.h>
#include "corosync/cfg.h"
#include "corosync/cpg.h"
#include "common.h"

#define PIDFILE "/var/run/clucpd.pid"

#define LOG(severity, format ...) do { if (glbl_debug) \
  printf(format); \
  else syslog(severity, format); \
} while (0)

#define _CS_WARN(func, params ...) \
  if ((error = func(params)) != CS_OK) { \
    LOG(LOG_WARNING, "%s() failed: %s (%i)\n", \
        #func, cs_error_to_string(error), error)

#define CS_WARN(func, params ...) do { \
  _CS_WARN(func, params); } \
} while (0)

#define CS_WARN_GOTO(target, func, params ...) do { \
  _CS_WARN(func, params); \
  goto target; } \
} while (0)

static int glbl_debug = 0;
static int glbl_running = 1;
static corosync_cfg_handle_t glbl_cfg_handle;
static unsigned int glbl_local_nodeid;

void cfg_state_track_callback (
  corosync_cfg_state_notification_buffer_t *notification_buffer,
  cs_error_t error)
{
  notification_buffer = notification_buffer;
  error = error;
}

void cfg_shutdown_callback (corosync_cfg_handle_t cfg_handle,
                            corosync_cfg_shutdown_flags_t flags)
{
  cfg_handle = cfg_handle;
  flags = flags;
}

void send_status (unsigned char opcode, cpg_handle_t handle, uint32_t nodeid,
                  char *addrs, char *filename, ssize_t filesize, char *err1,
                  char *err2, char *err3)
{
  struct iovec iov[6];
  int iovcnt;
  cs_error_t error;

  /* packet format (host byte order):
     opcode      (unsigned char)
     filename\0  (char[])
     if opcode == CS_OK: filesize    (ssize_t)
     else:               errortext\0 (char[])
   */
  iov[0].iov_base = &opcode;
  iov[0].iov_len = sizeof(opcode);
  iov[1].iov_base = filename;
  iov[1].iov_len = strlen(filename) + sizeof(char);

  if (opcode == OP_ERR) {
    LOG(LOG_WARNING, "%s%s from node %u (%s)%s%s\n",
        err1, filename, nodeid, addrs, err2, err3);

    iov[2].iov_base = err1;
    iov[2].iov_len = strlen(err1);
    iov[3].iov_base = filename;
    iov[3].iov_len = strlen(filename);
    iov[4].iov_base = err2;
    iov[4].iov_len = strlen(err2);
    iov[5].iov_base = err3;
    iov[5].iov_len = strlen(err3) + 1;
    iovcnt = 6;
  } else {
    iovcnt = 3;
    iov[2].iov_base = &filesize;
    iov[2].iov_len = sizeof(filesize);
  }

  CS_WARN(cpg_mcast_joined, handle, CPG_TYPE_AGREED, iov, iovcnt);
}

void cpg_deliver_callback (cpg_handle_t handle,
                           const struct cpg_name *group_name, uint32_t nodeid,
                           uint32_t pid, void *msg, size_t msg_len)
{
  struct header hdr;
  void *data;
  char *c;
  char *filename;
  int fd;
  ssize_t datalen;
  ssize_t ret;
  char addrs[NODE_ADDRESSES_BUFFER_SIZE];

  handle = handle;
  pid = pid;

  if (nodeid == glbl_local_nodeid)
    /* ignore message from local node */
    return;

  if (strncmp(GROUPNAME, group_name->value, group_name->length))
    /* not for us */
    return;

  if (msg_len < sizeof(hdr) + sizeof(char)) return;

  make_node_addresses(glbl_cfg_handle, nodeid, addrs, sizeof(addrs));

  /* packet format (host byte order):
     opcode      (unsigned char)
     filemode    (mode_t)
     owner       (uid_t)
     group       (gid_t)
     fullpath\0  (char[])
     data        (void[])
   */
  memcpy(&hdr, msg, sizeof(hdr));
  switch (hdr.opcode) {
    case OP_OK:
    case OP_ERR: return;
    case OP_PUSH: break;
    default:
      LOG(LOG_INFO, "unknown opcode %u from node %u (%s)\n",
          hdr.opcode, nodeid, addrs);
      return;
  }

  filename = msg + sizeof(hdr);
  for (c = filename; (void*) c < msg + msg_len; c++)
    if (*c == '\0') break;

  if ((void*) c >= msg + msg_len) {
    LOG(LOG_INFO, "message from node %u (%s) too short for filename\n",
        nodeid, addrs);
    return;
  }

  data = c + 1;
  datalen = msg_len - (data - msg);

  if ((fd = open(filename, O_WRONLY|O_CREAT|O_TRUNC, hdr.mode)) < 0) {
    send_status(OP_ERR, handle, nodeid, addrs, filename, datalen,
                "cannot open ", " for writing: ", strerror(errno));
    return;
  }

  if (fchown(fd, hdr.uid, hdr.gid) < 0) {
    send_status(OP_ERR, handle, nodeid, addrs, filename, datalen,
                "cannot chown ", ": ", strerror(errno));
    close(fd);
    return;
  }

  ret = write(fd, data, datalen);
  if (ret < 0) {
    send_status(OP_ERR, handle, nodeid, addrs, filename, datalen,
                "cannot write ", ": ", strerror(errno));
    close(fd);
    return;
  }
  if (ret != datalen) {
    send_status(OP_ERR, handle, nodeid, addrs, filename, datalen,
                "cannot write ", ": ", "couldn't write all data");
    close(fd);
    return;
  }

  if (close(fd) < 0) {
    send_status(OP_ERR, handle, nodeid, addrs, filename, datalen,
                "cannot close ", ": ", strerror(errno));
    return;
  }

  LOG(LOG_INFO, "got %s (%lu B) from node %u (%s)\n",
      filename, datalen, nodeid, addrs);

  send_status(OP_OK, handle, nodeid, addrs, filename, datalen,
              NULL, NULL, NULL);
}

void cpg_confchg_callback (cpg_handle_t handle,
                           const struct cpg_name *group_name,
                           const struct cpg_address *member_list,
                           size_t member_list_entries,
                           const struct cpg_address *left_list,
                           size_t left_list_entries,
                           const struct cpg_address *joined_list,
                           size_t joined_list_entries)
{
  handle = handle;
  group_name = group_name;
  member_list = member_list;
  member_list_entries = member_list_entries;
  left_list = left_list;
  left_list_entries = left_list_entries;
  joined_list = joined_list;
  joined_list_entries = joined_list_entries;
}

/* wait for action on corosync socket */
int wait_message (int fd)
{
  fd_set fds;
  struct timeval timeout;
  int res;

  for (;;) {
    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    timeout.tv_sec = 60;
    timeout.tv_usec = 0;

    res = select(fd + 1, &fds, NULL, NULL, &timeout);
    if (res < 0) {
      if (errno != EINTR) LOG(LOG_WARNING, "select(): %s\n", strerror(errno));
      break;
    }
    if ((res > 0) && FD_ISSET(fd, &fds)) break;
  }

  return res;
}

void main_loop ()
{
  cs_error_t error;
  cpg_handle_t cpg_handle;
  corosync_cfg_callbacks_t cfg_callbacks = {
    .corosync_cfg_state_track_callback = cfg_state_track_callback,
    .corosync_cfg_shutdown_callback = cfg_shutdown_callback,
  };
  cpg_callbacks_t cpg_callbacks = {
    .cpg_deliver_fn = cpg_deliver_callback,
    .cpg_confchg_fn = cpg_confchg_callback,
  };
  struct cpg_name group_name = {
    .length = strlen(GROUPNAME),
    .value = GROUPNAME,
  };
  int fd;

  CS_WARN_GOTO(ML_ERR_0, corosync_cfg_initialize, &glbl_cfg_handle,
                                                  &cfg_callbacks);
  CS_WARN_GOTO(ML_ERR_1, cpg_initialize, &cpg_handle, &cpg_callbacks);
  CS_WARN_GOTO(ML_ERR_2, cpg_join, cpg_handle, &group_name);
  CS_WARN_GOTO(ML_ERR_3, corosync_cfg_local_get, glbl_cfg_handle,
                                                 &glbl_local_nodeid);
  CS_WARN_GOTO(ML_ERR_3, cpg_fd_get, cpg_handle, &fd);

  while (wait_message(fd) > 0) {
    CS_WARN_GOTO(ML_ERR_3, cpg_dispatch, cpg_handle, CS_DISPATCH_ALL);
  }

ML_ERR_3:
  CS_WARN(cpg_leave, cpg_handle, &group_name);
ML_ERR_2:
  CS_WARN(cpg_finalize, cpg_handle);
ML_ERR_1:
  CS_WARN(corosync_cfg_finalize, glbl_cfg_handle);
ML_ERR_0:
  return;
}
 
void kill_handler (int sig)
{
  sig = sig;
  glbl_running = 0;
}

int setsighandler (int sig, void (*func)(int))
{
  struct sigaction sa;

  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = func;
  return sigaction(sig, &sa, NULL);
}

void print_usage ()
{
  puts("clucpd " VERSION "\n"
"usage: clucpd [-p <pidfile>] [-f] [-d]\n"
"       clucpd -h\n\n"
"  -p <pidfile>    save pid to <pidfile> instead of " PIDFILE "\n"
"  -f              run in foregroud, don't daemonize\n"
"  -d              debug mode, don't daemonize, don't save pid, "
                   "log to stdout\n"
"  -h              this help ;-)");
}

int main (int argc, char **argv)
{
  int i;
  int foreground = 0;
  char *pidfile = PIDFILE;
  FILE *f;
  pid_t pid;

  while ((i = getopt(argc, argv, "hfdp:")) != -1) {
    switch (i) {
      case 'h': print_usage(); return 0;
      case 'f': foreground = 1; break;
      case 'd': foreground = 1; glbl_debug = 1; break;
      case 'p': pidfile = optarg; break;
      default: fputs("for help use -h\n", stderr); return 1;
    }
  }

  if (chdir("/") < 0) err(1, "cannot chdir to /");

  if (!foreground) {
    pid = fork();
    if (pid < 0) err(1, "cannot fork");
    if (pid > 0) return 0;

    if (setsid() < 0) err(1, "setsid()");
  }

  if (!glbl_debug) {
    for (i = 0; i < 32; i++) setsighandler(i, SIG_IGN);
    if (setsighandler(SIGTERM, kill_handler) < 0)
      err(1, "cannot set signal handler for SIGTERM");

    if ((f = fopen(pidfile, "w")) == NULL)
      err(1, "cannot open %s for writing", pidfile);
    fprintf(f, "%i", getpid());
    fclose(f);

    for (i = 0; i <= 2; i++) close(i);
    openlog("clucpd", LOG_PID, LOG_DAEMON);
  }

  LOG(LOG_INFO, "starting...\n");

  for (;;) {
    main_loop();
    if (!glbl_running) break;

    /* maybe corosync exited, wait some seconds and reconnect */
    LOG(LOG_INFO, "sleeping for 15 seconds due to corosync error\n");
    sleep(15);
  }

  LOG(LOG_INFO, "exiting...\n");

  if (!foreground && (unlink(pidfile) < 0))
    LOG(LOG_WARNING, "cannot unlink %s: %s", pidfile, strerror(errno));

  if (!glbl_debug) closelog();

  return 0;
}
